/*
 Yaku v0.1.9
 (c) 2015 Yad Smood. http://ysmood.org
 License MIT
*/
(function(root) {
  var Yaku;
  return Yaku = (function() {

    /**
    	 * This class follows the [Promises/A+](https://promisesaplus.com) and
    	 * [ES6](http://people.mozilla.org/~jorendorff/es6-draft.html#sec-promise-objects) spec
    	 * with some extra helpers.
    	 * @param  {Function} executor Function object with two arguments resolve and reject.
    	 * The first argument fulfills the promise, the second argument rejects it.
    	 * We can call these functions, once our operation is completed.
    	 * @example
    	 * ```coffee
    	 * Promise = require 'yaku'
    	 * p = new Promise (resolve, reject) ->
    	 * 	setTimeout ->
    	 * 		if Math.random() > 0.5
    	 * 			resolve 'ok'
    	 * 		else
    	 * 			reject 'no'
    	 * ```
     */
    var $circularChain, $handlerStack, $hasUnhandledRejection, $invalid_argument, $noop, $pending, $prePromise, $rejected, $resolved, $settlerStack, $tryCatchFn, $tryErr, addHandler, assertIterable, callHanler, genScheduler, genSettler, genTraceInfo, genTryCatcher, genTypeError, getThen, isFunction, isLongStackTrace, isObject, newEmptyYaku, release, scheduleHandler, scheduleUnhandledRejection, settlePromise, settleWithX, settleXthen, tryCatcher;

    function Yaku(executor) {
      var err;
      if (isLongStackTrace) {
        this[$handlerStack] = genTraceInfo();
      }
      if (executor === $noop) {
        return;
      }
      err = genTryCatcher(executor)(genSettler(this, $resolved), genSettler(this, $rejected));
      if (err === $tryErr) {
        settlePromise(this, $rejected, err.e);
      }
    }


    /**
    	 * Appends fulfillment and rejection handlers to the promise,
    	 * and returns a new promise resolving to the return value of the called handler.
    	 * @param  {Function} onFulfilled Optional. Called when the Promise is resolved.
    	 * @param  {Function} onRejected  Optional. Called when the Promise is rejected.
    	 * @return {Yaku} It will return a new Yaku which will resolve or reject after
    	 * @example
    	 * the current Promise.
    	 * ```coffee
    	 * Promise = require 'yaku'
    	 * p = Promise.resolve 10
    	 *
    	 * p.then (v) ->
    	 * 	console.log v
    	 * ```
     */

    Yaku.prototype.then = function(onFulfilled, onRejected) {
      return addHandler(this, newEmptyYaku(), onFulfilled, onRejected);
    };


    /**
    	 * The `catch()` method returns a Promise and deals with rejected cases only.
    	 * It behaves the same as calling `Promise.prototype.then(undefined, onRejected)`.
    	 * @param  {Function} onRejected A Function called when the Promise is rejected.
    	 * This function has one argument, the rejection reason.
    	 * @return {Yaku} A Promise that deals with rejected cases only.
    	 * @example
    	 * ```coffee
    	 * Promise = require 'yaku'
    	 * p = Promise.reject 10
    	 *
    	 * p.catch (v) ->
    	 * 	console.log v
    	 * ```
     */

    Yaku.prototype["catch"] = function(onRejected) {
      return this.then(void 0, onRejected);
    };


    /**
    	 * The `Promise.resolve(value)` method returns a Promise object that is resolved with the given value.
    	 * If the value is a thenable (i.e. has a then method), the returned promise will "follow" that thenable,
    	 * adopting its eventual state; otherwise the returned promise will be fulfilled with the value.
    	 * @param  {Any} value Argument to be resolved by this Promise.
    	 * Can also be a Promise or a thenable to resolve.
    	 * @return {Yaku}
    	 * @example
    	 * ```coffee
    	 * Promise = require 'yaku'
    	 * p = Promise.resolve 10
    	 * ```
     */

    Yaku.resolve = function(value) {
      if (value instanceof Yaku) {
        return value;
      }
      return settleWithX(newEmptyYaku(), value);
    };


    /**
    	 * The `Promise.reject(reason)` method returns a Promise object that is rejected with the given reason.
    	 * @param  {Any} reason Reason why this Promise rejected.
    	 * @return {Yaku}
    	 * @example
    	 * ```coffee
    	 * Promise = require 'yaku'
    	 * p = Promise.reject 10
    	 * ```
     */

    Yaku.reject = function(reason) {
      return settlePromise(newEmptyYaku(), $rejected, reason);
    };


    /**
    	 * The `Promise.race(iterable)` method returns a promise that resolves or rejects
    	 * as soon as one of the promises in the iterable resolves or rejects,
    	 * with the value or reason from that promise.
    	 * @param  {iterable} iterable An iterable object, such as an Array.
    	 * @return {Yaku} The race function returns a Promise that is settled
    	 * the same way as the first passed promise to settle.
    	 * It resolves or rejects, whichever happens first.
    	 * @example
    	 * ```coffee
    	 * Promise = require 'yaku'
    	 * Promise.race [
    	 * 	123
    	 * 	Promise.resolve 0
    	 * ]
    	 * .then (value) ->
    	 * 	console.log value # => 123
    	 * ```
     */

    Yaku.race = function(iterable) {
      var i, len, p;
      assertIterable(iterable);
      len = iterable.length;
      if (len === 0) {
        return Yaku.resolve([]);
      }
      p = newEmptyYaku();
      i = 0;
      while (i < len) {
        settleWithX(p, iterable[i++]);
        if (p._state !== $pending) {
          break;
        }
      }
      return p;
    };


    /**
    	 * The `Promise.all(iterable)` method returns a promise that resolves when
    	 * all of the promises in the iterable argument have resolved.
    	 *
    	 * The result is passed as an array of values from all the promises.
    	 * If something passed in the iterable array is not a promise,
    	 * it's converted to one by Promise.resolve. If any of the passed in promises rejects,
    	 * the all Promise immediately rejects with the value of the promise that rejected,
    	 * discarding all the other promises whether or not they have resolved.
    	 * @param  {iterable} iterable An iterable object, such as an Array.
    	 * @return {Yaku}
    	 * @example
    	 * ```coffee
    	 * Promise = require 'yaku'
    	 * Promise.all [
    	 * 	123
    	 * 	Promise.resolve 0
    	 * ]
    	 * .then (values) ->
    	 * 	console.log values # => [123, 0]
    	 * ```
     */

    Yaku.all = function(iterable) {
      var convertor, countDown, i, iter, len, onRejected, p1, res;
      assertIterable(iterable);
      convertor = Yaku.resolve;
      len = countDown = iterable.length;
      if (len === 0) {
        return convertor([]);
      }
      p1 = newEmptyYaku();
      res = [];
      i = 0;
      onRejected = function(reason) {
        settlePromise(p1, $rejected, reason);
      };
      iter = function(i) {
        convertor(iterable[i]).then(function(value) {
          res[i] = value;
          if (!--countDown) {
            settlePromise(p1, $resolved, res);
          }
        }, onRejected);
      };
      while (i < len) {
        iter(i++);
      }
      return p1;
    };


    /**
    	 * Catch all possibly unhandled rejections. If you want to use specific
    	 * format to display the error stack, overwrite it.
    	 * If it is set, auto `console.error` unhandled rejection will be disabed.
    	 * @param {Any} reason The rejection reason.
    	 * @example
    	 * ```coffee
    	 * Promise = require 'yaku'
    	 * Promise.onUnhandledRejection = (reason) ->
    	 * 	console.error reason
    	 *
    	 * # The console will log an unhandled rejection error message.
    	 * Promise.reject('my reason')
    	 *
    	 * # The below won't log the unhandled rejection error message.
    	 * Promise.reject('v').catch ->
    	 * ```
     */

    Yaku.onUnhandledRejection = function(reason, p) {
      var $info, hStack, stack;
      if (!isObject(console)) {
        return;
      }
      $info = 'Unhandled Rejection:';
      hStack = '';
      if (isLongStackTrace && p[$handlerStack]) {
        while (p) {
          hStack = p[$handlerStack] + hStack;
          p = p[$prePromise];
        }
      }
      if (typeof __filename === 'string') {
        hStack = hStack.replace(RegExp(".+" + __filename + ".+\\n", "g"), '');
      }
      stack = reason ? reason.stack || reason : reason;
      return console.error($info, stack, hStack);
    };

    isLongStackTrace = false;


    /**
    	 * It is used to enable the long stack trace.
    	 * Once it is enabled, it can't be reverted.
    	 * @example
    	 * ```coffee
    	 * Promise = require 'yaku'
    	 * Promise.enableLongStackTrace()
    	 * ```
     */

    Yaku.enableLongStackTrace = function() {
      isLongStackTrace = true;
    };


    /*
    	 * All static variable name will begin with `$`. Such as `$rejected`.
    	 * @private
     */

    $tryCatchFn = null;

    $tryErr = {
      e: null
    };

    $noop = {};

    isObject = function(obj) {
      return typeof obj === 'object';
    };

    isFunction = function(obj) {
      return typeof obj === 'function';
    };


    /**
    	 * Release the specified key of an object.
    	 * @private
    	 * @param  {Object} obj
    	 * @param  {String | Number} key
     */

    release = function(obj, key) {
      obj[key] = void 0;
    };


    /**
    	 * Wrap a function into a try-catch.
    	 * @private
    	 * @return {Any | $tryErr}
     */

    tryCatcher = function() {
      var e;
      try {
        return $tryCatchFn.apply(this, arguments);
      } catch (_error) {
        e = _error;
        $tryErr.e = e;
        return $tryErr;
      }
    };


    /**
    	 * Generate a try-catch wrapped function.
    	 * @private
    	 * @param  {Function} fn
    	 * @return {Function}
     */

    genTryCatcher = function(fn) {
      $tryCatchFn = fn;
      return tryCatcher;
    };


    /**
    	 * Generate a scheduler.
    	 * @private
    	 * @param  {Integer}  initQueueSize
    	 * @param  {Function} fn `(Yaku, Value) ->` The schedule handler.
    	 * @return {Function} `(Yaku, Value) ->` The scheduler.
     */

    genScheduler = function(initQueueSize, fn) {

      /**
      		 * All async promise will be scheduled in
      		 * here, so that they can be execute on the next tick.
      		 * @private
       */
      var flush, fnQueue, fnQueueLen, scheduleFlush;
      fnQueue = Array(initQueueSize);
      fnQueueLen = 0;

      /**
      		 * Run all queued functions.
      		 * @private
       */
      flush = function() {
        var i, p, pIndex, v, vIndex;
        i = 0;
        while (i < fnQueueLen) {
          pIndex = i++;
          vIndex = i++;
          p = fnQueue[pIndex];
          v = fnQueue[vIndex];
          release(fnQueue, pIndex);
          release(fnQueue, vIndex);
          fn(p, v);
        }
        fnQueueLen = 0;
        fnQueue.length = initQueueSize;
      };

      /**
      		 * Schedule a flush task on the next tick.
      		 * @private
      		 * @param {Function} fn The flush task.
       */
      scheduleFlush = (function() {
        var content, doc, mutationObserver, nextTick, node, observer;
        doc = root.document;
        try {
          nextTick = process.nextTick;
          return function() {
            nextTick(flush);
          };
        } catch (_error) {}
        if (nextTick = root.setImmediate) {
          return function() {
            nextTick(flush);
          };
        } else if (mutationObserver = root.MutationObserver) {
          content = 1;
          node = doc.createTextNode('');
          observer = new mutationObserver(flush);
          observer.observe(node, {
            characterData: true
          });
          return function() {
            node.data = (content = -content);
          };
        } else {
          return function() {
            setTimeout(flush);
          };
        }
      })();
      return function(p, v) {
        fnQueue[fnQueueLen++] = p;
        fnQueue[fnQueueLen++] = v;
        if (fnQueueLen === 2) {
          scheduleFlush();
        }
      };
    };


    /**
    	 * Check if a variable is an iterable object.
    	 * @private
    	 * @param  {Any}  obj
    	 * @return {Boolean}
     */

    assertIterable = function(obj) {
      if (obj instanceof Array) {
        return;
      }
      throw genTypeError($invalid_argument);
    };


    /**
    	 * Generate type error object.
    	 * @private
    	 * @param  {String} msg
    	 * @return {TypeError}
     */

    genTypeError = function(msg) {
      return new TypeError(msg);
    };

    genTraceInfo = function() {
      return (new Error).stack.replace('Error', 'From previous event:');
    };


    /**
    	 * These are some static symbolys.
    	 * @private
     */

    $rejected = 0;

    $resolved = 1;

    $pending = 2;

    $hasUnhandledRejection = '_isUnhandled';

    $prePromise = '_pre';

    $handlerStack = '_hStack';

    $settlerStack = '_sStack';

    $circularChain = 'promise_circular_chain';

    $invalid_argument = 'invalid_argument';

    Yaku.prototype._state = $pending;


    /**
    	 * The number of current promises that attach to this Yaku instance.
    	 * @private
     */

    Yaku.prototype._pCount = 0;

    Yaku.prototype["" + $hasUnhandledRejection] = true;


    /**
    	 * Create an empty promise.
    	 * @private
    	 * @return {Yaku}
     */

    newEmptyYaku = function() {
      return new Yaku($noop);
    };


    /**
    	 * It will produce a settlePromise function to user.
    	 * Such as the resolve and reject in this `new Yaku (resolve, reject) ->`.
    	 * @private
    	 * @param  {Yaku} self
    	 * @param  {Integer} state The value is one of `$pending`, `$resolved` or `$rejected`.
    	 * @return {Function} `(value) -> undefined` A resolve or reject function.
     */

    genSettler = function(self, state) {
      return function(value) {
        return settlePromise(self, state, value);
      };
    };


    /**
    	 * Link the promise1 to the promise2.
    	 * @private
    	 * @param {Yaku} p1
    	 * @param {Yaku} p2
    	 * @param {Function} onFulfilled
    	 * @param {Function} onRejected
     */

    addHandler = function(p1, p2, onFulfilled, onRejected) {
      if (isFunction(onFulfilled)) {
        p2._onFulfilled = onFulfilled;
      }
      if (isFunction(onRejected)) {
        p2._onRejected = onRejected;
      }
      p1[$hasUnhandledRejection] = false;
      if (p1._state === $pending) {
        p1[p1._pCount++] = p2;
      } else {
        scheduleHandler(p1, p2);
      }
      return p2;
    };


    /**
    	 * Resolve the value returned by onFulfilled or onRejected.
    	 * @private
    	 * @param {Yaku} p1
    	 * @param {Yaku} p2
     */

    scheduleHandler = genScheduler(1000, function(p1, p2) {
      var handler, x;
      handler = p1._state ? p2._onFulfilled : p2._onRejected;
      if (handler === void 0) {
        settlePromise(p2, p1._state, p1._value);
        return;
      }
      x = genTryCatcher(callHanler)(handler, p1._value);
      if (x === $tryErr) {
        settlePromise(p2, $rejected, x.e);
        return;
      }
      settleWithX(p2, x);
    });

    scheduleUnhandledRejection = genScheduler(100, function(p) {
      if (p[$hasUnhandledRejection]) {
        Yaku.onUnhandledRejection(p._value, p);
      }
    });

    callHanler = function(handler, value) {
      return handler(value);
    };


    /**
    	 * Resolve or reject a promise.
    	 * @private
    	 * @param  {Yaku} p
    	 * @param  {Integer} state
    	 * @param  {Any} value
     */

    settlePromise = function(p, state, value) {
      var i, len;
      if (p._state !== $pending) {
        return;
      }
      p._state = state;
      p._value = value;
      i = 0;
      len = p._pCount;
      if (state === $rejected && !len) {
        if (isLongStackTrace) {
          p[$settlerStack] = genTraceInfo();
        }
        scheduleUnhandledRejection(p);
      }
      while (i < len) {
        scheduleHandler(p, p[i]);
        release(p, i++);
      }
      return p;
    };


    /**
    	 * Resolve or reject primise with value x. The x can also be a thenable.
    	 * @private
    	 * @param {Yaku} p
    	 * @param {Any | Thenable} x A normal value or a thenable.
     */

    settleWithX = function(p, x) {
      var xthen;
      if (x === p && x) {
        settlePromise(p, $rejected, genTypeError($circularChain));
        return;
      }
      if (x !== null && (isFunction(x) || isObject(x))) {
        xthen = genTryCatcher(getThen)(x);
        if (xthen === $tryErr) {
          settlePromise(p, $rejected, xthen.e);
          return;
        }
        if (isFunction(xthen)) {
          if (isLongStackTrace) {
            p[$prePromise] = x;
          }
          settleXthen(p, x, xthen);
        } else {
          settlePromise(p, $resolved, x);
        }
      } else {
        settlePromise(p, $resolved, x);
      }
      return p;
    };


    /**
    	 * Try to get a promise's then method.
    	 * @private
    	 * @param  {Thenable} x
    	 * @return {Function}
     */

    getThen = function(x) {
      return x.then;
    };


    /**
    	 * Resolve then with its promise.
    	 * @private
    	 * @param  {Yaku} p
    	 * @param  {Thenable} x
    	 * @param  {Function} xthen
     */

    settleXthen = function(p, x, xthen) {
      var err;
      err = genTryCatcher(xthen).call(x, function(y) {
        if (!x) {
          return;
        }
        x = null;
        settleWithX(p, y);
      }, function(r) {
        if (!x) {
          return;
        }
        x = null;
        settlePromise(p, $rejected, r);
      });
      if (err === $tryErr && x) {
        settlePromise(p, $rejected, err.e);
        x = null;
      }
    };

    try {
      module.exports = Yaku;
    } catch (_error) {
      try {
        define(function() {
          return Yaku;
        });
      } catch (_error) {
        window.Yaku = Yaku;
      }
    }

    return Yaku;

  })();
})(this);
